import logging

import pytest
import full_match
from loguru import logger as loguru_logger

from emptylog import LoggersGroup, MemoryLogger


def test_len_of_group():
    assert len(LoggersGroup()) == 0
    assert len(LoggersGroup(MemoryLogger())) == 1
    assert len(LoggersGroup(MemoryLogger(), MemoryLogger())) == 2
    assert len(LoggersGroup(MemoryLogger(), MemoryLogger(), MemoryLogger())) == 3
    assert len(LoggersGroup(MemoryLogger(), MemoryLogger(), MemoryLogger(), MemoryLogger())) == 4

    assert len(LoggersGroup() + LoggersGroup()) == 0
    assert len(LoggersGroup() + MemoryLogger()) == 1
    assert len(MemoryLogger() + MemoryLogger()) == 2
    assert len(MemoryLogger() + MemoryLogger() + MemoryLogger()) == 3


@pytest.mark.parametrize(
    ['wrong_logger', 'exception_message'],
    (
        (1, 'A logger group can only be created from loggers. You passed 1 (int).'),
        ('kek', 'A logger group can only be created from loggers. You passed \'kek\' (str).'),
        (None, 'A logger group can only be created from loggers. You passed None (NoneType).'),
    ),
)
def test_create_group_with_not_loggers(wrong_logger, exception_message):
    with pytest.raises(TypeError, match=full_match(exception_message)):
        LoggersGroup(wrong_logger)


@pytest.mark.parametrize(
    ['get_method'],
    (
        (lambda x: x.debug,),
        (lambda x: x.info,),
        (lambda x: x.warning,),
        (lambda x: x.error,),
        (lambda x: x.exception,),
        (lambda x: x.critical,),
    ),
)
def test_run_group_of_memory_loggers(get_method):
    first_internal_logger = MemoryLogger()
    second_internal_logger = MemoryLogger()
    group = LoggersGroup(first_internal_logger, second_internal_logger)

    get_method(group)('lol', 'kek', cheburek='pek')

    for internal_logger in first_internal_logger, second_internal_logger:
        assert len(get_method(internal_logger.data)) == 1
        assert get_method(internal_logger.data)[0].message == 'lol'
        assert get_method(internal_logger.data)[0].args == ('kek',)
        assert get_method(internal_logger.data)[0].kwargs == {'cheburek': 'pek'}


def test_repr_loggers_group():
    assert repr(LoggersGroup()) == 'LoggersGroup()'
    assert repr(LoggersGroup(LoggersGroup())) == 'LoggersGroup(LoggersGroup())'
    assert repr(LoggersGroup(MemoryLogger())) == 'LoggersGroup(MemoryLogger())'
    assert repr(LoggersGroup(MemoryLogger(), MemoryLogger())) == 'LoggersGroup(MemoryLogger(), MemoryLogger())'
    assert repr(LoggersGroup(MemoryLogger(), MemoryLogger(), MemoryLogger())) == 'LoggersGroup(MemoryLogger(), MemoryLogger(), MemoryLogger())'


def test_empty_group_plus_empty_group():
    assert type(LoggersGroup() + LoggersGroup()) is LoggersGroup
    assert (LoggersGroup() + LoggersGroup()).loggers == ()


def test_not_empty_group_plus_empty_group():
    first_internal_logger = MemoryLogger()
    second_internal_logger = MemoryLogger()

    assert type(LoggersGroup(first_internal_logger) + LoggersGroup()) is LoggersGroup
    assert type(LoggersGroup(first_internal_logger, second_internal_logger) + LoggersGroup()) is LoggersGroup

    assert len((LoggersGroup(first_internal_logger) + LoggersGroup()).loggers) == 1
    assert len((LoggersGroup(first_internal_logger, second_internal_logger) + LoggersGroup()).loggers) == 2

    assert (LoggersGroup(first_internal_logger) + LoggersGroup()).loggers[0] is first_internal_logger
    assert (LoggersGroup(first_internal_logger, second_internal_logger) + LoggersGroup()).loggers[0] is first_internal_logger
    assert (LoggersGroup(first_internal_logger, second_internal_logger) + LoggersGroup()).loggers[1] is second_internal_logger


def test_empty_group_plus_not_empty_group():
    first_internal_logger = MemoryLogger()
    second_internal_logger = MemoryLogger()

    assert type(LoggersGroup() + LoggersGroup(first_internal_logger)) is LoggersGroup
    assert type(LoggersGroup() + LoggersGroup(first_internal_logger, second_internal_logger)) is LoggersGroup

    assert len((LoggersGroup() + LoggersGroup(first_internal_logger)).loggers) == 1
    assert len((LoggersGroup() + LoggersGroup(first_internal_logger, second_internal_logger)).loggers) == 2

    assert (LoggersGroup() + LoggersGroup(first_internal_logger)).loggers[0] is first_internal_logger
    assert (LoggersGroup() + LoggersGroup(first_internal_logger, second_internal_logger)).loggers[0] is first_internal_logger
    assert (LoggersGroup() + LoggersGroup(first_internal_logger, second_internal_logger)).loggers[1] is second_internal_logger


def test_empty_group_plus_another_logger():
    another_logger = MemoryLogger()

    assert type(LoggersGroup() + another_logger) is LoggersGroup
    assert len((LoggersGroup() + another_logger).loggers) == 1
    assert (LoggersGroup() + another_logger).loggers[0] is another_logger


def test_another_logger_plus_empty_group():
    another_logger = MemoryLogger()

    assert type(another_logger + LoggersGroup()) is LoggersGroup
    assert len((another_logger + LoggersGroup()).loggers) == 1
    assert len(another_logger + LoggersGroup()) == 1
    assert (another_logger + LoggersGroup()).loggers[0] is another_logger


@pytest.mark.parametrize(
    ['third_party_logger'],
    (
        (loguru_logger,),
        (logging,),
        (logging.getLogger('kek'),),
    ),
)
def test_empty_group_plus_third_party_logger(third_party_logger):
    first_group = LoggersGroup()

    sum = first_group + third_party_logger

    assert type(sum) is LoggersGroup
    assert sum is not first_group
    assert len(sum.loggers) == 1
    assert len(sum) == 1
    assert sum.loggers[0] is third_party_logger


@pytest.mark.parametrize(
    ['third_party_logger'],
    (
        (loguru_logger,),
        (logging,),
        (logging.getLogger('kek'),),
    ),
)
def test_third_party_logger_plus_empty_group(third_party_logger):
    first_group = LoggersGroup()

    sum = third_party_logger + first_group

    assert type(sum) is LoggersGroup
    assert sum is not first_group
    assert len(sum.loggers) == 1
    assert len(sum) == 1
    assert sum.loggers[0] is third_party_logger


@pytest.mark.parametrize(
    ['loggers'],
    (
        ([loguru_logger, logging, logging.getLogger('kek')],),
        ([MemoryLogger(), MemoryLogger()],),
        ([MemoryLogger()],),
        ([],),
    ),
)
def test_iteration_by_group(loggers):
    group = LoggersGroup(*loggers)

    assert loggers == [x for x in group]



import logging

from loguru import logger as loguru_logger

from emptylog import LoggerProtocol, EmptyLogger, LoggersGroup, MemoryLogger, PrintingLogger


def test_positive_examples_of_runtime_check():
    assert isinstance(logging, LoggerProtocol)
    assert isinstance(logging.getLogger('kek'), LoggerProtocol)
    assert isinstance(logging.getLogger('lol'), LoggerProtocol)
    assert isinstance(EmptyLogger(), LoggerProtocol)
    assert isinstance(LoggersGroup(), LoggerProtocol)
    assert isinstance(MemoryLogger(), LoggerProtocol)
    assert isinstance(PrintingLogger(), LoggerProtocol)


def test_negative_examples_of_runtime_check():
    assert not isinstance(1, LoggerProtocol)
    assert not isinstance('logging', LoggerProtocol)


def test_loguru_logger_is_logger():
    assert isinstance(loguru_logger, LoggerProtocol)



from emptylog.call_data import LoggerCallData
from emptylog.accumulated_data import LoggerAccumulatedData


def test_fill_accumulated_data_and_check_size():
    data = LoggerAccumulatedData()

    lists = [
        data.debug,
        data.info,
        data.warning,
        data.error,
        data.exception,
        data.critical,
    ]

    logs_sum = 0

    for index, logs_list in enumerate(lists):
        assert len(data) == logs_sum

        for _ in range(index + 1):
            logs_list.append(LoggerCallData('some message', (), {}))
            logs_sum += 1

            assert len(data) == logs_sum

    assert len(data) == 21



from emptylog.empty_logger import EmptyLogger as EmptyLogger  # noqa: F401
from emptylog.protocols import LoggerProtocol as LoggerProtocol  # noqa: F401
from emptylog.memory_logger import MemoryLogger as MemoryLogger  # noqa: F401
from emptylog.printing_logger import PrintingLogger as PrintingLogger  # noqa: F401
from emptylog.loggers_group import LoggersGroup as LoggersGroup  # noqa: F401



import sys
from typing import Tuple, Callable, Any
from threading import Lock
from collections.abc import Iterator

from printo import descript_data_object

from emptylog.protocols import LoggerProtocol, LoggerMethodProtocol
from emptylog.abstract_logger import AbstractLogger


if sys.version_info < (3, 9):
    GroupIterator = Iterator  # pragma: no cover
else:
    GroupIterator = Iterator[LoggerProtocol]  # pragma: no cover

class LoggersGroup(AbstractLogger):
    loggers: Tuple[LoggerProtocol, ...]

    def __init__(self, *loggers: LoggerProtocol) -> None:
        for logger in loggers:
            if not isinstance(logger, LoggerProtocol):
                raise TypeError(f'A logger group can only be created from loggers. You passed {repr(logger)} ({type(logger).__name__}).')

        self.loggers = loggers
        self.lock = Lock()

    def __repr__(self) -> str:
        return descript_data_object(type(self).__name__, self.loggers, {})

    def __len__(self) -> int:
        return len(self.loggers)

    def __iter__(self) -> GroupIterator:  # type: ignore[type-arg]
        yield from self.loggers

    def debug(self, message: str, *args: Any, **kwargs: Any) -> None:
        self.run_loggers(lambda x: x.debug, message, *args, **kwargs)

    def info(self, message: str, *args: Any, **kwargs: Any) -> None:
        self.run_loggers(lambda x: x.info, message, *args, **kwargs)

    def warning(self, message: str, *args: Any, **kwargs: Any) -> None:
        self.run_loggers(lambda x: x.warning, message, *args, **kwargs)

    def error(self, message: str, *args: Any, **kwargs: Any) -> None:
        self.run_loggers(lambda x: x.error, message, *args, **kwargs)

    def exception(self, message: str, *args: Any, **kwargs: Any) -> None:
        self.run_loggers(lambda x: x.exception, message, *args, **kwargs)

    def critical(self, message: str, *args: Any, **kwargs: Any) -> None:
        self.run_loggers(lambda x: x.critical, message, *args, **kwargs)

    def run_loggers(self, get_method: Callable[[LoggerProtocol], LoggerMethodProtocol], message: str, *args: Any, **kwargs: Any) -> None:
        with self.lock:
            for logger in self.loggers:
                method = get_method(logger)
                method(message, *args, **kwargs)
